import jax.random
import numpy as np
import pytest
from jax import numpy as jnp

from jaxns.internals.interp_utils import get_interp_indices_and_weights, apply_interp, left_broadcast_multiply, \
    InterpolatedArray


def test_get_interp_indices_and_weights():
    xp = [0, 1, 2, 3]
    x = 1.5
    (i0, alpha0), (i1, alpha1) = get_interp_indices_and_weights(x, xp)
    print(i0, alpha0, i1, alpha1)
    assert i0 == 1
    assert alpha0 == 0.5
    assert i1 == 2
    assert alpha1 == 0.5

    x = 0
    (i0, alpha0), (i1, alpha1) = get_interp_indices_and_weights(x, xp)
    print(i0, alpha0, i1, alpha1)
    assert i0 == 0
    assert alpha0 == 1
    assert i1 == 1
    assert alpha1 == 0

    x = 3
    (i0, alpha0), (i1, alpha1) = get_interp_indices_and_weights(x, xp)
    print(i0, alpha0, i1, alpha1)
    assert i0 == 2
    assert alpha0 == 0
    assert i1 == 3
    assert alpha1 == 1

    x = -1
    (i0, alpha0), (i1, alpha1) = get_interp_indices_and_weights(x, xp)
    print(i0, alpha0, i1, alpha1)
    assert xp[i0] * alpha0 + xp[i1] * alpha1 == -1

    x = 4
    (i0, alpha0), (i1, alpha1) = get_interp_indices_and_weights(x, xp)
    print(i0, alpha0, i1, alpha1)
    assert xp[i0] * alpha0 + xp[i1] * alpha1 == 4

    x = 5
    (i0, alpha0), (i1, alpha1) = get_interp_indices_and_weights(x, xp)
    print(i0, alpha0, i1, alpha1)
    assert xp[i0] * alpha0 + xp[i1] * alpha1 == 5

    xp = [0., 0.]
    x = 0.
    (i0, alpha0), (i1, alpha1) = get_interp_indices_and_weights(x, xp)
    print(i0, alpha0, i1, alpha1)
    assert i0 == 0
    assert alpha0 == 1
    assert i1 == 1
    assert alpha1 == 0.

    xp = [0., 0.]
    x = -1
    (i0, alpha0), (i1, alpha1) = get_interp_indices_and_weights(x, xp)
    print(i0, alpha0, i1, alpha1)
    assert i0 == 0
    assert alpha0 == 2.
    assert i1 == 1
    assert alpha1 == -1.

    xp = [0.]
    x = 0.5
    (i0, alpha0), (i1, alpha1) = get_interp_indices_and_weights(x, xp)
    print(i0, alpha0, i1, alpha1)
    assert i0 == 0
    assert alpha0 == 1.
    assert i1 == 0
    assert alpha1 == 0.

    # Vector ops
    xp = [0, 1, 2, 3]
    x = [1.5, 1.5]
    (i0, alpha0), (i1, alpha1) = get_interp_indices_and_weights(x, xp)
    print(i0, alpha0, i1, alpha1)
    np.testing.assert_array_equal(i0, jnp.asarray([1, 1]))
    np.testing.assert_array_equal(alpha0, jnp.asarray([0.5, 0.5]))
    np.testing.assert_array_equal(i1, jnp.asarray([2, 2]))
    np.testing.assert_array_equal(alpha1, jnp.asarray([0.5, 0.5]))

    xp = [0, 1, 2, 3]
    x = [1.5, 2.5]
    (i0, alpha0), (i1, alpha1) = get_interp_indices_and_weights(x, xp)
    print(i0, alpha0, i1, alpha1)
    np.testing.assert_array_equal(i0, jnp.asarray([1, 2]))
    np.testing.assert_array_equal(alpha0, jnp.asarray([0.5, 0.5]))
    np.testing.assert_array_equal(i1, jnp.asarray([2, 3]))
    np.testing.assert_array_equal(alpha1, jnp.asarray([0.5, 0.5]))

    # xp = [0, 1, 2, 3]
    # x = [-0.5, 3.5]
    # (i0, alpha0), (i1, alpha1) = get_interp_indices_and_weights(x, xp)
    # print(i0, alpha0, i1, alpha1)
    # np.testing.assert_array_equal(i0, jnp.asarray([1, 2]))
    # np.testing.assert_array_equal(alpha0, jnp.asarray([0.5, 0.5]))
    # np.testing.assert_array_equal(i1, jnp.asarray([2, 3]))
    # np.testing.assert_array_equal(alpha1, jnp.asarray([0.5, 0.5]))


@pytest.mark.parametrize('regular_grid', [True, False])
def test_apply_interp(regular_grid):
    xp = jnp.linspace(0., 1., 10)
    x = jnp.linspace(-0.1, 1.1, 10)
    (i0, alpha0), (i1, alpha1) = get_interp_indices_and_weights(x, xp, regular_grid=regular_grid)
    np.testing.assert_allclose(apply_interp(xp, i0, alpha0, i1, alpha1), x, atol=1e-6)

    x = jnp.linspace(-0.1, 1.1, 10)
    (i0, alpha0), (i1, alpha1) = get_interp_indices_and_weights(x, xp, regular_grid=regular_grid)
    assert apply_interp(jnp.zeros((4, 5, 10, 6)), i0, alpha0, i1, alpha1, axis=2).shape == (4, 5, 10, 6)

    x = 0.
    (i0, alpha0), (i1, alpha1) = get_interp_indices_and_weights(x, xp, regular_grid=regular_grid)
    assert apply_interp(jnp.zeros((4, 5, 10, 6)), i0, alpha0, i1, alpha1, axis=2).shape == (4, 5, 6)

    print(
        jax.jit(lambda x: apply_interp(x, i0, alpha0, i1, alpha1, axis=2)).lower(
            jnp.zeros((4, 5, 10, 6))).compile().cost_analysis()
    )
    # [{'bytes accessed': 1440.0, 'utilization1{}': 2.0, 'bytes accessed0{}': 960.0, 'bytes accessedout{}': 480.0, 'bytes accessed1{}': 960.0}]
    # [{'bytes accessed1{}': 960.0,  'utilization1{}': 2.0, 'bytes accessedout{}': 480.0, 'bytes accessed0{}': 960.0, 'bytes accessed': 1440.0}]


def test_regular_grid():
    # Inside bounds
    xp = jnp.linspace(0., 1., 10)
    fp = jax.random.normal(jax.random.PRNGKey(0), (10, 15))
    x = jnp.linspace(0., 1., 100)
    (i0, alpha0), (i1, alpha1) = get_interp_indices_and_weights(x, xp, regular_grid=False)
    f_no = apply_interp(fp, i0, alpha0, i1, alpha1)

    (i0, alpha0), (i1, alpha1) = get_interp_indices_and_weights(x, xp, regular_grid=True)
    f_yes = apply_interp(fp, i0, alpha0, i1, alpha1)
    np.testing.assert_allclose(
        f_yes, f_no,
        atol=1e-6
    )

    # Outside bounds
    x = jnp.linspace(-0.1, 1.1, 100)
    xp = jnp.linspace(0., 1., 10)
    fp = jax.random.normal(jax.random.PRNGKey(0), (10, 15))
    (i0, alpha0), (i1, alpha1) = get_interp_indices_and_weights(x, xp, regular_grid=False)
    f_no = apply_interp(fp, i0, alpha0, i1, alpha1)

    (i0, alpha0), (i1, alpha1) = get_interp_indices_and_weights(x, xp, regular_grid=True)
    f_yes = apply_interp(fp, i0, alpha0, i1, alpha1)
    np.testing.assert_allclose(
        f_yes, f_no,
        atol=1e-6
    )


def test_left_broadcast_multiply():
    assert np.all(left_broadcast_multiply(np.ones((2, 3)), np.ones((2,))) == np.ones((2, 3)))
    assert np.all(left_broadcast_multiply(np.ones((2, 3)), np.ones((2, 3))) == np.ones((2, 3)))
    assert np.all(
        left_broadcast_multiply(np.ones((1, 2, 3, 4, 5)), np.ones((3, 4)), axis=2) == np.ones((1, 2, 3, 4, 5)))
    assert np.all(
        left_broadcast_multiply(np.ones((1, 2, 3, 4, 5)), np.ones((3, 4)), axis=-3) == np.ones((1, 2, 3, 4, 5)))


@pytest.mark.parametrize('regular_grid', [True, False])
def test_interpolated_array(regular_grid: bool):
    # scalar time
    times = jnp.linspace(0, 10, 100)
    values = jnp.sin(times)
    interp = InterpolatedArray(times, values, regular_grid=regular_grid)
    assert interp(5.).shape == ()
    np.testing.assert_allclose(interp(5.), jnp.sin(5), atol=2e-3)

    # vector time
    assert interp(jnp.array([5., 6.])).shape == (2,)
    np.testing.assert_allclose(interp(jnp.array([5., 6.])), jnp.sin(jnp.array([5., 6.])), atol=2e-3)

    # Now with axis = 1
    times = jnp.linspace(0, 10, 100)
    values = jnp.stack([jnp.sin(times), jnp.cos(times)], axis=0)  # [2, 100]
    interp = InterpolatedArray(times, values, axis=1, regular_grid=regular_grid)
    assert interp(5.).shape == (2,)
    np.testing.assert_allclose(interp(5.), jnp.array([jnp.sin(5), jnp.cos(5)]), atol=2e-3)

    # Vector
    assert interp(jnp.array([5., 6., 7.])).shape == (2, 3)
    np.testing.assert_allclose(interp(jnp.array([5., 6., 7.])),
                               jnp.stack([jnp.sin(jnp.array([5., 6., 7.])), jnp.cos(jnp.array([5., 6., 7.]))],
                                         axis=0),
                               atol=2e-3)



from typing import NamedTuple

import jax
import numpy as np
from jax import random, numpy as jnp

from jaxns.plotting import weighted_percentile
from jaxns.utils import resample, _bit_mask, save_pytree, load_pytree


def test_resample():
    x = random.normal(key=random.PRNGKey(0), shape=(50,))
    logits = -jnp.ones(50)
    samples = {'x': x}
    assert jnp.all(resample(random.PRNGKey(0), samples, logits)['x'] == resample(random.PRNGKey(0), x, logits))


def test_bit_mask():
    assert _bit_mask(1, width=2) == [1, 0]
    assert _bit_mask(2, width=2) == [0, 1]
    assert _bit_mask(3, width=2) == [1, 1]


def test_weighted_percentile():
    # Test the weighted percentile function
    samples = np.asarray([1, 2, 3, 4, 5])
    log_weights = np.asarray([0, 0, 0, 0, 0])
    percentiles = [50]
    assert np.allclose(weighted_percentile(samples, log_weights, percentiles), 3.0)


class MockPyTree171(NamedTuple):
    x: jax.Array


def test_gh171(tmp_path):
    pytree = MockPyTree171(jnp.array([1., 2., 3.]))
    save_pytree(pytree, str(tmp_path / "results.json"))
    loaded_pytree = load_pytree(str(tmp_path / "results.json"))
    np.testing.assert_allclose(loaded_pytree.x, pytree.x)



import time
from functools import partial
from typing import NamedTuple

import jax
import numpy as np
import pytest
from jax import numpy as jnp
from tensorflow_probability.substrates import jax as tfp

from jaxns.internals.prefix_sum import _slice_along_axis, _interleave, scan_associative


def test__slice_along_axis():
    a = jnp.arange(5)
    np.testing.assert_allclose(
        _slice_along_axis(a, stop=-1),
        a[:-1]
    )


def test__interleave():
    # even
    a = jnp.asarray([0, 1, 2, 3])
    b = jnp.asarray([4, 5, 6, 7])
    expected_interleaved = jnp.asarray([0, 4, 1, 5, 2, 6, 3, 7])
    np.testing.assert_allclose(_interleave(a, b, axis=0), expected_interleaved)

    # odd
    a = jnp.asarray([0, 1, 2])
    b = jnp.asarray([3, 4, 5])
    expected_interleaved = jnp.asarray([0, 3, 1, 4, 2, 5])
    np.testing.assert_allclose(_interleave(a, b, axis=0), expected_interleaved)

    # multi-dim
    a = jnp.asarray(
        [
            [0, 1, 2],
            [0, 1, 2]
        ]
    )
    b = jnp.asarray(
        [
            [3, 4, 5],
            [3, 4, 5]
        ]
    )
    expected_interleaved_1 = jnp.asarray(
        [
            [0, 3, 1, 4, 2, 5],
            [0, 3, 1, 4, 2, 5]
        ]
    )
    expected_interleaved_0 = jnp.asarray(
        [
            [0, 1, 2],
            [3, 4, 5],
            [0, 1, 2],
            [3, 4, 5]
        ]
    )
    np.testing.assert_allclose(_interleave(a, b, axis=1), expected_interleaved_1)
    np.testing.assert_allclose(_interleave(a, b, axis=0), expected_interleaved_0)


@pytest.mark.parametrize('num_elems', [10, 11])
@pytest.mark.parametrize('op', [jnp.add, jnp.multiply, jnp.minimum, jnp.maximum])
def test_scan_associative(num_elems, op):
    elems = jnp.arange(num_elems)
    results = scan_associative(
        op,
        elems,
        axis=0
    )
    np.testing.assert_allclose(
        results,
        tfp.math.scan_associative(
            op,
            elems
        )
    )


@pytest.mark.parametrize('num_elems', [10, 11])
def test_scan_associative_pytee(num_elems):
    class MockClass(NamedTuple):
        x: jax.Array
        y: jax.Array

    def op(x: MockClass, y: MockClass) -> MockClass:
        return MockClass(
            x=x.x + y.x,
            y=x.y + y.y
        )

    elems = MockClass(
        x=jnp.arange(num_elems),
        y=jnp.arange(num_elems)
    )
    results = scan_associative(
        op,
        elems,
        axis=0
    )
    np.testing.assert_allclose(
        results,
        tfp.math.scan_associative(
            op,
            elems
        )
    )


def test_scan_associative_rank_reducing():
    def per_elem_op(x) -> jax.Array:
        return jnp.sum(x)

    def associative_op(x, y):
        print(x.shape, y.shape)
        assert np.shape(x) == np.shape(x)
        assert np.shape(x) == ()
        return per_elem_op(x) + per_elem_op(y)

    xs = jnp.arange(10)
    _ = scan_associative(associative_op, xs)


@pytest.mark.parametrize('N', [32, 128, 512])
@pytest.mark.parametrize('M', [32, 128, 512])
def test_performance(N: int, M: int):
    import tensorflow_probability.substrates.jax as tfp
    inputs = jnp.ones((N, M, M))
    op = jnp.linalg.matmul

    fn_compiled = jax.jit(partial(tfp.math.scan_associative, fn=op)).lower(elems=inputs).compile()

    m = 3

    t0 = time.time()
    for _ in range(m):
        fn_compiled(elems=inputs).block_until_ready()
    t1 = time.time()
    tfp_runtime = (t1 - t0) / m

    fn_compiled = jax.jit(partial(scan_associative, fn=op)).lower(elems=inputs).compile()

    t0 = time.time()
    for _ in range(m):
        fn_compiled(elems=inputs).block_until_ready()
    t1 = time.time()
    internal_runtime = (t1 - t0) / m
    print(f"sequence length {N}, matrix mul {M}x{M} @ {M}x{M}:")
    print(f"\tRun time tfp.math.scan_associative {tfp_runtime}")
    print(f"\tRun time internal scan_associative {internal_runtime}")
    if internal_runtime < tfp_runtime:
        print(f"Internal faster by {100. * (1. - internal_runtime / tfp_runtime)} %")
    else:
        print(f"tfp.math.scan_associative faster by {100. * (1. - tfp_runtime / internal_runtime)} %")



import dataclasses
from typing import Union, Tuple

import jax
import jax.numpy as jnp
import numpy as np
from jax import lax


def apply_interp(x: jax.Array, i0: jax.Array, alpha0: jax.Array, i1: jax.Array, alpha1: jax, axis: int = 0):
    """
    Apply interpolation alpha given axis.

    Args:
        x: nd-array
        i0: [N] or scalar
        alpha0: [N] or scalar
        i1: [N] or scalar
        alpha1: [N] or scalar
        axis: axis to take along

    Returns:
        [N] or scalar interpolated along axis
    """

    def take(i):
        num_dims = len(np.shape(x))
        # [0] [1] [2 3 4], num_dims=5, axis=1
        slices = [slice(None)] * axis + [i] + [slice(None)] * (num_dims - axis - 1)
        # return jnp.take(x, i, axis=axis)
        return x[tuple(slices)]

    return left_broadcast_multiply(take(i0), alpha0, axis=axis) + left_broadcast_multiply(
        take(i1), alpha1, axis=axis)


def left_broadcast_multiply(x, y, axis: int = 0):
    """
    Left broadcast multiply of two arrays.
    Equivalent to right-padding before multiply

    Args:
        x: [..., a,b,c,...]
        y: [a, b]

    Returns:
        [..., a, b, c, ...]
    """
    needed_length = len(np.shape(x)[axis:])
    len_y = len(np.shape(y))
    extra = needed_length - len_y
    if extra < 0:
        raise ValueError(f"Shape mismatch {np.shape(x)} x {np.shape(y)}.")
    y = lax.reshape(y, np.shape(y) + (1,) * extra)
    return x * y


def get_interp_indices_and_weights(x, xp, regular_grid: bool = False) -> Tuple[
    Tuple[Union[int, jax.Array, float, jax.Array]], Tuple[Union[int, jax.Array, float, jax.Array]]]:
    """
    One-dimensional linear interpolation. Outside bounds is also linear from nearest two points.

    Args:
        x: the x-coordinates at which to evaluate the interpolated values
        xp: the x-coordinates of the data points, must be increasing

    Returns:
        the interpolated values, same shape as `x`
    """

    x = jnp.asarray(x, dtype=jnp.float_)
    xp = jnp.asarray(xp, dtype=jnp.float_)
    if len(np.shape(xp)) == 0:
        xp = jnp.reshape(xp, (-1,))
    if np.shape(xp)[0] == 0:
        raise ValueError("xp must be non-empty")
    if np.shape(xp)[0] == 1:
        return (jnp.zeros_like(x, dtype=jnp.int32), jnp.ones_like(x)), (
            jnp.zeros_like(x, dtype=jnp.int32), jnp.zeros_like(x))

    # Find xp[i1-1] < x <= xp[i1]
    if regular_grid:
        # Use faster index determination
        delta_x = xp[1] - xp[0]
        i1 = jnp.clip((jnp.ceil((x - xp[0]) / delta_x)).astype(jnp.int64), 1, len(xp) - 1)
        i0 = i1 - 1
    else:
        i1 = jnp.clip(jnp.searchsorted(xp, x, side='right'), 1, len(xp) - 1)
        i0 = i1 - 1

    dx = xp[i1] - xp[i0]
    delta = x - xp[i0]

    epsilon = np.spacing(np.finfo(xp.dtype).eps)
    dx0 = jnp.abs(dx) <= epsilon  # Prevent NaN gradients when `dx` is small.
    dx = jnp.where(dx0, 1, dx)
    alpha = delta / dx
    return (i0, (1. - alpha)), (i1, alpha)


@dataclasses.dataclass(eq=False)
class InterpolatedArray:
    x: jax.Array  # [N]
    values: jax.Array  # [..., N, ...] `axis` has N elements

    axis: int = 0
    regular_grid: bool = False

    def __post_init__(self):

        if len(np.shape(self.x)) != 1:
            raise ValueError(f"Times must be 1D, got {np.shape(self.x)}.")

        def _assert_shape(x):
            if np.shape(x)[self.axis] != np.size(self.x):
                raise ValueError(f"Input values must have time length on `axis` dimension, got {np.shape(x)}.")

        jax.tree.map(_assert_shape, self.values)

        self.x, self.values = jax.tree.map(jnp.asarray, (self.x, self.values))

    @property
    def shape(self):
        return jax.tree.map(lambda x: np.shape(x)[:self.axis] + np.shape(x)[self.axis + 1:], self.values)

    def __call__(self, time: jax.Array) -> jax.Array:
        """
        Interpolate at time based on input times.

        Args:
            time: time to evaluate at.

        Returns:
            value at given time
        """
        (i0, alpha0), (i1, alpha1) = get_interp_indices_and_weights(time, self.x, regular_grid=self.regular_grid)
        return jax.tree.map(lambda x: apply_interp(x, i0, alpha0, i1, alpha1, axis=self.axis), self.values)



import functools
from typing import TypeVar, Callable

import jax
import jax.numpy as jnp
import numpy as np
from jax import lax


def _slice_along_axis(x, start=0, stop=None, step=1, axis=0):
    """Slices a Tensor along the given axis."""

    def _slice(x):
        if axis >= 0:
            slices = [slice(None)] * axis + [slice(start, stop, step)]
        else:
            slices = [Ellipsis, slice(start, stop, step)] + [slice(None)] * (-1 - axis)
        return x[tuple(slices)]

    return jax.tree.map(_slice, x)


def _interleave(a, b, axis):
    """Interleaves two `Tensor`s along the given axis."""

    # [a b c ...] [d e f ...] -> [a d b e c f ...]

    def _op(a, b):
        if np.shape(a) != np.shape(b):
            raise ValueError('`a` and `b` must have the same shape.')

        stacked = jnp.stack([a, b], axis=axis + 1)

        output_shape = np.shape(a)[:axis] + (2 * np.shape(a)[axis],) + np.shape(a)[axis + 1:]
        return lax.reshape(stacked, output_shape)

    return jax.tree.map(_op, a, b)


X = TypeVar('X')


def scan_associative(fn: Callable[[X, X], X], elems: X, axis: int = 0) -> X:
    """
    Perform a scan with an associative binary operation, in parallel.

    Suitable for fn: (X, X) -> X where (f o f) o f == f o (f o f)

    The associative scan operation computes the cumulative sum, or
    [all-prefix sum](https://en.wikipedia.org/wiki/Prefix_sum), of a set of
    elements under an associative binary operation [1]. For example, using the
    ordinary addition operator `fn = lambda a, b: a + b`, this is equivalent to
    the ordinary cumulative sum.

    The associative structure allows the computation to be decomposed
    and executed by parallel reduction. Where a naive sequential
    implementation would loop over all `N` elements, this method requires
    only a logarithmic number (`2 * ceil(log_2 N)`) of sequential steps, and
    can thus yield substantial performance speedups from hardware-accelerated
    vectorization. The total number of invocations of the binary operation
    (including those performed in parallel) is
    `2 * (N / 2 + N / 4 + ... + 1) = 2N - 2`
    --- i.e., approximately twice as many as a naive approach.

    [1] Blelloch, Guy E.
        [Prefix sums and their applications](
        https://www.cs.cmu.edu/~guyb/papers/Ble93.pdf)
        Technical Report CMU-CS-90-190,
        School of Computer Science,
        Carnegie Mellon University, 1990.

    Args:
        fn: the associative binary operation to perform.
        elems: [..., N, ...] the input elements to scan over.
        axis: the axis to scan over.

    Returns:
        [..., N, ...] cumulative operation applied on input.
    """

    # This is a rewrite of tfp.math.scan_associative, which relies on broadcasting, to better handle vectorisation.
    # This tends to be more performant for smaller ops,
    # and TFP's version tends to be more performant for larger ops by about 10%.
    # Equivalent performance for 100 matmuls of 100x100 matrices
    # (See test_performance for more on performance.)

    _num_elements = np.shape(jax.tree.leaves(elems)[0])[axis]
    if _num_elements < 2:
        raise ValueError(f"Must be at least 2 elements, got {_num_elements}.")

    odd_slice = functools.partial(_slice_along_axis, start=1, stop=None, step=2, axis=axis)
    even_slice = functools.partial(_slice_along_axis, start=0, stop=None, step=2, axis=axis)
    trim_slice = functools.partial(_slice_along_axis, start=0, stop=-1, step=1, axis=axis)
    pretrim_slice = functools.partial(_slice_along_axis, start=1, stop=None, step=1, axis=axis)
    zero_slice = functools.partial(_slice_along_axis, start=0, stop=1, step=1, axis=axis)
    last_slice = functools.partial(_slice_along_axis, start=-1, stop=None, step=1, axis=axis)

    # The basic idea of the algorithm is as follows, using "sum" to represent the associative operation:
    # 1. The cumulative sum is: y0=x0, y1=x0+x1, y2=x0+x1+x2, y3=x0+x1+x2+x3
    # 2. Thus odd indices are equivalent to a two part process:
    #       a. Sum even+odd inputs: (x0+x1), (x2+x3)
    #       b. recursively apply `_scan` on the result to give: y1=x0+x1, y3=x0+x1+x2+x3
    # 3. The even indices are then found by:
    #       a. adding even indices (ignoring x0 in this step): y2=y1+x2
    #       b. preconcatenating x0: y0=x0
    # 4. The result of a single `_scan` is then interleaving the above: _interleave([y0, y1, y2], [y1, y2, y3])

    # TODO: If the associative identity, `id`, was provided then we could combine 3b into 3a via: y0=id+x0

    def _scan(elems):
        """Perform scan on `elems`."""
        # Compute _scan(op(elem_odds, elem_evens)) -> odd indicies
        num_elements = np.shape(jax.tree.leaves(elems)[0])[axis]
        if num_elements == 0:
            raise RuntimeError(f"Got zero length `elems`")
        elif num_elements == 1:
            return elems
        elif num_elements == 2:
            elems_0 = jax.tree.map(zero_slice, elems)
            elems_1 = jax.tree.map(last_slice, elems)
            consecutive_sum = jax.vmap(fn, in_axes=axis, out_axes=axis)(
                elems_0,
                elems_1
            )  # x0+x1
            _results = [
                elems_0,  # x0
                consecutive_sum  # x0+x1
            ]
            results = jax.tree.map(
                lambda *x: jnp.concatenate(x, axis=axis),
                *_results
            )
            return results

        is_even = num_elements % 2 == 0
        if is_even:
            # Want odd number of elements
            last_elem = jax.tree.map(last_slice, elems)
            elems = jax.tree.map(lambda *x: jnp.concatenate(x, axis=axis), elems, last_elem)
            # elems = jax.tree.map(trim_slice, elems)  # x0, x1, x2, x3, x4, x5, x6 (+ x7)
        elems_odd = jax.tree.map(odd_slice, elems)  # x1, x3, x5
        elems_even = jax.tree.map(even_slice, elems)  # x0, x2, x4, x6
        consecutive_sum = jax.vmap(fn, in_axes=axis, out_axes=axis)(
            jax.tree.map(trim_slice, elems_even),  # x0, x2, x4
            elems_odd  # x1, x3, x5
        )  # x0+x1, x2+x3, x4+x5
        result_odd = _scan(consecutive_sum)  # x0+x1, x0+x1+x2+x3, x0+x1+x2+x3+x4+x5

        result_even = jax.vmap(fn, in_axes=axis, out_axes=axis)(
            result_odd,
            jax.tree.map(pretrim_slice, elems_even)
        )  # x0+x1+x2, x0+x1+x2+x3+x4, x0+x1+x2+x3+x4+x5+x6

        partial_results = _interleave(
            result_odd,
            result_even,
            axis=axis
        )
        # x0+x1, x0+x1+x2,
        # x0+x1+x2+x3, x0+x1+x2+x3+x4,
        # x0+x1+x2+x3+x4+x5, x0+x1+x2+x3+x4+x5+x6

        _results = [
            jax.tree.map(zero_slice, elems),  # x0
            partial_results
        ]

        results = jax.tree.map(
            lambda *x: jnp.concatenate(x, axis=axis),
            *_results
        )

        if is_even:
            results = jax.tree.map(trim_slice, results)

        return results

    result = _scan(elems)

    return result



import jax
import jax.numpy as jnp
from jax import lax


def quick_unit(x: jax.Array) -> jax.Array:
    """
    Quick approximation to the sigmoid.

    Args:
        x: jax.Array value in (-inf, inf) open interval

    Returns:
        value in (0, 1) in open interval
    """
    return 0.5 * (x / (1 + lax.abs(x)) + 1)


def quick_unit_inverse(y: jax.Array) -> jax.Array:
    """
    Inverse of quick_unit.

    Args:
        y: jax.Array value in (0, 1) open interval

    Returns:
        value in (-inf, inf) in open interval
    """
    twoy = y + y

    return jnp.where(
        y >= 0.5,
        (1 - twoy) / (twoy - 2),
        1 - lax.reciprocal(twoy)
    )



import warnings
from typing import Callable

import jax
import numpy as np

from jaxns.internals.types import float_type, LikelihoodType

__all__ = [
    'jaxify_likelihood'
]

def jaxify_likelihood(log_likelihood: Callable[..., np.ndarray], vectorised: bool = False) -> LikelihoodType:
    """
    Wraps a non-JAX log likelihood function.

    Args:
        log_likelihood: a non-JAX log-likelihood function, which accepts a number of arguments and returns a scalar
            log-likelihood.
        vectorised: if True then the `log_likelihood` performs a vectorised computation for leading batch dimensions,
            i.e. if a leading batch dimension is added to all input arguments, then it returns a vector of
            log-likelihoods with the same leading batch dimension.

    Returns:
        A JAX-compatible log-likelihood function.
    """
    warnings.warn(
        "You're using a non-JAX log-likelihood function. This may be slower than a JAX log-likelihood function. "
        "Also, you are responsible for ensuring that the function is deterministic. "
        "Also, you cannot use learnable parameters in the likelihood call."
    )

    def _casted_log_likelihood(*args) -> np.ndarray:
        return np.asarray(log_likelihood(*args), dtype=float_type)

    def _log_likelihood(*args) -> jax.Array:
        # Define the expected shape & dtype of output.
        result_shape_dtype = jax.ShapeDtypeStruct(
            shape=(),
            dtype=float_type
        )
        return jax.pure_callback(_casted_log_likelihood, result_shape_dtype, *args, vectorized=vectorised)

    return _log_likelihood



import warnings
from typing import Optional, List, Union

import jax
import jax.numpy as jnp
import numpy as np
import pylab as plt
from jax import random
from matplotlib.animation import FuncAnimation
from mpl_toolkits.axes_grid1 import make_axes_locatable
from scipy.stats import gaussian_kde

from jaxns.internals.log_semiring import cumulative_logsumexp, LogSpace, normalise_log_space
from jaxns.internals.shapes import tuple_prod
from jaxns.internals.types import NestedSamplerResults, int_type
from jaxns.utils import resample

__all__ = ['plot_diagnostics',
           'plot_cornerplot']


def plot_diagnostics(results: NestedSamplerResults, save_name=None):
    """
    Plot diagnostics of the nested sampling run.

    Args:
        results: NestedSamplerResult
        save_name: file to save figure to.
    """

    num_samples = int(results.total_num_samples)
    if results.log_L_samples.shape[0] != num_samples:
        raise ValueError(f"Expected all samples to have the same number of samples, "
                         f"got log_L_samples with {results.log_L_samples.shape[0]} samples, "
                         f"expected {num_samples} samples.")
    fig, axs = plt.subplots(5, 1, sharex=True, figsize=(8, 12))
    log_X = np.asarray(results.log_X_mean)
    num_live_points_per_sample = np.asarray(results.num_live_points_per_sample)
    log_L = np.asarray(results.log_L_samples)
    max_log_likelihood = np.max(log_L)
    log_dp_mean = np.asarray(results.log_dp_mean)
    log_cum_evidence = cumulative_logsumexp(log_dp_mean)
    cum_evidence = np.exp(log_cum_evidence)
    log_Z_mean = np.asarray(results.log_Z_mean)
    num_likelihood_evaluations_per_sample = np.asarray(results.num_likelihood_evaluations_per_sample)
    if np.any(num_likelihood_evaluations_per_sample == 0):
        warnings.warn("Found samples with zero likelihood evaluations.")
        efficiency = np.where(
            num_likelihood_evaluations_per_sample == 0,
            np.nan,
            1. / num_likelihood_evaluations_per_sample
        )
    else:
        efficiency = 1. / num_likelihood_evaluations_per_sample
    mean_efficiency = np.exp(results.log_efficiency)
    # Plot the number of live points
    axs[0].plot(-log_X, num_live_points_per_sample, c='black')
    axs[0].set_ylabel(r'$n_{\rm live}$')
    # detect if too small log likelihood
    rel_log_L = log_L - max_log_likelihood
    axs[1].plot(-log_X, np.exp(rel_log_L), c='black')
    axs[1].axhline(1., color='black', ls='dashed',
                   label=rf"$\log L_{{\rm max}}={max_log_likelihood:.1f}$")
    axs[1].set_ylabel(r'$L/L_{\rm max}$')
    axs[1].legend()
    axs[2].plot(-log_X, np.exp(log_dp_mean), c='black')
    axs[2].axvline(-results.H_mean, color='black', ls='dashed',
                   label=rf'$-H={-results.H_mean:.1f}$')
    axs[2].set_ylabel(r'$Z^{-1}L dX$')
    axs[2].legend()
    axs[3].plot(-log_X, cum_evidence, c='black')
    axs[3].axhline(1., color='black', ls='dashed',
                   label=rf"$\log Z={log_Z_mean:.1f}$")
    axs[3].set_ylabel(r'$Z(\lambda > L)/Z$')
    axs[3].legend()
    axs[4].scatter(-log_X, efficiency, s=2, c='black')

    axs[4].axhline(mean_efficiency, color='black', ls='dashed',
                   label=f'avg. eff.={mean_efficiency:.3f}')
    axs[4].set_ylabel("sampler efficiency")
    axs[4].set_ylim(0., 1.05)
    axs[4].legend()
    axs[4].set_xlabel(r'$- \log X$')
    if save_name is not None:
        fig.savefig(save_name, bbox_inches='tight', dpi=300, pad_inches=0.0)
    plt.show()


def plot_cornerplot(results: NestedSamplerResults, variables: Optional[List[str]] = None,
                    with_parametrised: bool = False,
                    save_name: Optional[str] = None, kde_overlay: bool = False):
    """
    Plots a cornerplot of the posterior samples.

    Args:
        results: NestedSamplerResult
        variables: list of variable names to plot. Plots all collected samples by default.
        save_name: file to save result to.
        kde_overlay: whether to overlay a KDE on the histograms.
    """
    samples = results.samples
    if with_parametrised:
        samples.update(results.parametrised_samples)
    # Plot all variables by default
    if variables is None:
        variables = list(samples.keys())
    variables = sorted(filter(lambda v: v in samples, variables))
    ndims = sum([tuple_prod(samples[key].shape[1:]) for key in variables], 0)

    num_samples = int(results.total_num_samples)
    for key in variables:
        if samples[key].shape[0] != num_samples:
            raise ValueError(f"Expected all samples to have the same number of samples, "
                             f"got {key} with {samples[key].shape[0]} samples, "
                             f"expected {num_samples} samples.")

    # Get the leaves of the tree, and concatenate into [num_samples, ndims] shape
    leaves = np.concatenate(
        [np.asarray(samples[key]).reshape((num_samples, -1)) for key in variables],
        axis=-1
    )

    # Create a parameter for each column. For scalar parameters, we just use the name of the parameter.
    # For vector we use name[i,j,...] etc.
    parameters = []
    for key in variables:
        shape = samples[key].shape[1:]
        if tuple_prod(shape) == 1:
            parameters.append(key)
        else:
            # Loop over each dimension of the parameter, and create a parameter for each index
            for i in range(tuple_prod(shape)):
                indices = np.unravel_index(i, shape)
                parameters.append(f"{key}[{','.join([str(j) for j in indices])}]")

    # Get the maximum likelihood and MAP samples
    log_L_samples = np.asarray(results.log_L_samples)
    log_posterior_density = np.asarray(results.log_posterior_density)
    max_like_idx = np.argmax(log_L_samples)
    map_idx = np.argmax(log_posterior_density)
    max_like_sample = leaves[max_like_idx]
    map_sample = leaves[map_idx]

    # Get the weight of each sample
    log_weights = np.asarray(normalise_log_space(LogSpace(results.log_dp_mean), norm_type='max').log_abs_val)

    figsize = min(20, max(4, int(2 * ndims)))
    fig, axs = plt.subplots(ndims, ndims, figsize=(figsize, figsize), squeeze=False)

    # Get the number of bins for the histograms based on the effective sample size
    nbins = max(10, int(jnp.sqrt(results.ESS)))

    # Loop over the variables, and plot the marginal distributions on the diagonal setting a title above
    # each plot with the mean+-stddev, 5%/50%/95%, and MAP
    param_limits = dict()  # Store the 1_per and 99_per for each parameter
    for row in range(ndims):
        for col in range(ndims):
            ax = axs[row][col]
            if row != col:  # i == j ==> plot the marginal distribution
                continue
            # Plot the marginal distribution
            _samples = leaves[:, row]  # [num_samples]
            _parameter = parameters[row]
            _log_weights = log_weights
            is_finite = np.isfinite(_samples)
            if np.bitwise_not(np.all(is_finite)):
                warnings.warn(f"Found {np.sum(np.bitwise_not(is_finite))} non-finite samples for {_parameter}")
                _samples = _samples[is_finite]
                _log_weights = _log_weights[is_finite]
            _weights = np.exp(_log_weights)
            # Percentiles
            per_1, per_5, per_50, per_95, per_99 = weighted_percentile(_samples, _log_weights,
                                                                       [1, 5, 50, 95, 99])
            # Plot the histogram, from 1_per to 99_per
            ax.hist(_samples, bins=nbins, fc='None', edgecolor='black', density=True, weights=_weights,
                    range=(per_1, per_99))
            # Plot the maximum likelihood and MAP samples
            ax.axvline(max_like_sample[row], color='green')
            ax.axvline(map_sample[row], color='red')
            # Plot the mean and standard deviation
            sample_mean = np.average(_samples, weights=_weights)
            sample_std = np.sqrt(np.average((_samples - sample_mean) ** 2, weights=_weights))
            ax.axvline(sample_mean, linestyle='dashed', color='red')
            ax.axvline(sample_mean + sample_std, linestyle='dotted', color='red')
            ax.axvline(sample_mean - sample_std, linestyle='dotted', color='red')

            # Set the title
            title = [
                rf"${per_50:.2f}_{{{per_5:.2f}}}^{{{per_95:.2f}}}$",
                rf"${sample_mean:.2f}\pm{sample_std:.2f}$",
                rf"MAP ${map_sample[row]:.2f}$ | ML ${max_like_sample[row]:.2f}$"
            ]
            ax.set_title("\n".join(title))
            # Set the limits to 1 to 99 percentiles
            ax.set_xlim(per_1, per_99)
            param_limits[_parameter] = (per_1, per_99)

    # Plot the 2D histograms on lower-diagonal.
    for row in range(ndims):
        for col in range(ndims):
            ax = axs[row][col]
            if col >= row:
                continue

            # Get the samples for the 2D histogram
            _samples = leaves[:, [row, col]]  # [num_samples, 2]
            _log_weights = log_weights
            is_finite = np.all(np.isfinite(_samples), axis=-1)  # [num_samples]
            if np.bitwise_not(np.all(is_finite)):
                warnings.warn(
                    f"Found {np.sum(np.bitwise_not(is_finite))} non-finite samples for {parameters[row]} and {parameters[col]}")
                _samples = _samples[is_finite]
                _log_weights = _log_weights[is_finite]
            _weights = np.exp(_log_weights)

            # Plot the 2D histogram, over ranges set by the 1_per and 99_per of each parameter
            ranges = [param_limits[parameters[col]], param_limits[parameters[row]]]
            ax.hist2d(_samples[:, 1], _samples[:, 0], bins=(nbins, nbins), density=True,
                      cmap="bone_r",
                      weights=_weights, range=ranges)

            if kde_overlay:  # Put KDE contour on the 2D histograms

                # Calculate the point density
                x = _samples[:, 1]
                y = _samples[:, 0]
                xy = np.vstack([x, y])

                x_array = np.linspace(*param_limits[parameters[col]], 128)
                y_array = np.linspace(*param_limits[parameters[row]], 128)
                X, Y = np.meshgrid(x_array, y_array)
                xy_eval = np.vstack([X.ravel(), Y.ravel()])

                z = gaussian_kde(xy, weights=_weights)(xy_eval)
                z = z.reshape(X.shape)
                ax.contour(X, Y, z, levels=6, alpha=0.5)

            # Plot the maximum likelihood and MAP samples
            ax.scatter(max_like_sample[col], max_like_sample[row], color='green', marker='x')
            ax.scatter(map_sample[col], map_sample[row], color='red', marker='x')

            # Set the limits to 1 to 99 percentiles
            ax.set_xlim(param_limits[parameters[col]])
            ax.set_ylim(param_limits[parameters[row]])
    # Remove spacing
    plt.subplots_adjust(wspace=0.0, hspace=0.0)
    # Remove x ticks for all but bottom row
    for row in range(ndims - 1):
        for col in range(ndims):
            axs[row][col].set_xticks([])
            axs[row][col].set_xticklabels([])
    # Remove y ticks for all but left column
    for row in range(ndims):
        for col in range(1, ndims):
            axs[row][col].set_yticks([])
            axs[row][col].set_yticklabels([])
    # Set the labels on the bottom row and left column
    for i in range(ndims):
        axs[-1][i].set_xlabel(parameters[i])
        axs[i][0].set_ylabel(parameters[i])
    # Remove upper diagonal
    for row in range(ndims):
        for col in range(ndims):
            if col <= row:
                continue
            axs[row][col].remove()
    # Save the figure
    if save_name is not None:
        fig.savefig(save_name, bbox_inches='tight', dpi=300, pad_inches=0.0)
    plt.show()


def weighted_percentile(samples: np.ndarray, log_weights: np.ndarray,
                        percentiles: List[Union[float, int]]) -> np.ndarray:
    """
    Compute weighted percentiles of a set of samples.

    Args:
        samples: weighted samples
        log_weights: log weights of samples
        percentiles: list of percentiles to compute

    Returns:
        weighted percentiles
    """
    if len(percentiles) == 0:
        raise ValueError("percentiles must be a non-empty list")
    # Convert log weights to actual weights
    weights = LogSpace(log_weights - np.max(log_weights))  # Subtract max to avoid overflow
    weights = normalise_log_space(weights, norm_type='sum')  # Normalize weights

    # Sort samples and weights
    sorted_indices = np.argsort(samples)
    sorted_samples = samples[sorted_indices]
    sorted_weights = weights[sorted_indices]

    # Compute cumulative weights
    cumulative_weights = sorted_weights.cumsum()
    cumulative_weights = cumulative_weights - cumulative_weights[0]
    cumulative_weights = cumulative_weights / cumulative_weights[-1]
    # Add zero to start of cumulative weights

    # Compute weighted percentiles
    percentile_values = np.interp(np.asarray(percentiles) / 100.0, cumulative_weights.value, sorted_samples)
    return percentile_values


def plot_samples_development(results, variables=None, save_name=None):
    """
    Animate the live points in a corner plot, visualising how the algorithm proceeds.
    Caution, this can be very slow as it plots a frame per sample.

    Args:
        results: NestedSamplingResult
        vars: list of variable names to plot, or None
        save_name: '.mp4' file to save animation to.
    """
    if save_name is None:
        raise ValueError("In order to plot the animation we must save it.")
    # Plot all variables by default
    if variables is None:
        variables = list(results.samples.keys())
    variables = sorted(filter(lambda v: v in results.samples, variables))
    ndims = sum([tuple_prod(results.samples[key].shape[1:]) for key in variables], 0)
    figsize = min(20, max(4, int(2 * ndims)))
    fig, axs = plt.subplots(ndims, ndims, figsize=(figsize, figsize))
    if ndims == 1:
        axs = [[axs]]
    weights = jnp.exp(results.log_p_mean)
    max_samples = weights.size
    norm = plt.Normalize(weights.min(), weights.max())
    to_colour = lambda w: plt.cm.jet(norm(w))

    def _get_artists(artists, start, stop):
        lims = {}
        dim = 0
        for key in variables:  # sorted(results.samples.keys()):
            n1 = tuple_prod(results.samples[key].shape[1:])
            for i in range(n1):
                samples1 = results.samples[key].reshape((max_samples, -1))[:, i]
                samples1 = samples1[start:stop]
                dim2 = 0
                for key2 in variables:  # sorted(results.samples.keys()):
                    n2 = tuple_prod(results.samples[key2].shape[1:])
                    for i2 in range(n2):
                        ax = axs[dim][dim2]
                        if dim2 > dim:
                            dim2 += 1
                            ax.set_xticks([])
                            ax.set_xticklabels([])
                            ax.set_yticks([])
                            ax.set_yticklabels([])
                            continue
                        if n2 > 1:
                            title2 = "{}[{}]".format(key2, i2)
                        else:
                            title2 = "{}".format(key2)
                        if n1 > 1:
                            title1 = "{}[{}]".format(key, i)
                        else:
                            title1 = "{}".format(key)
                        # ax.set_title('{} {}'.format(title1, title2))
                        if dim == dim2:
                            _, _, new_patches = ax.hist(samples1)
                            artists = artists + list(new_patches)
                            lims[dim] = ax.get_xlim()
                        else:
                            samples2 = results.samples[key2].reshape((max_samples, -1))[:, i2]
                            samples2 = samples2[start:stop]

                            sc = ax.scatter(samples2, samples1, marker='+', c=to_colour(weights[start:stop]), alpha=0.3)
                            artists.append(sc)
                        if dim == ndims - 1:
                            ax.set_xlabel("{}".format(title2))
                        if dim2 == 0:
                            ax.set_ylabel("{}".format(title1))

                        dim2 += 1
                dim += 1
        for dim in range(ndims):
            for dim2 in range(ndims):
                if dim == dim2:
                    continue
                ax = axs[dim][dim2] if ndims > 1 else axs[0]
                if dim in lims.keys():
                    ax.set_ylim(lims[dim])
                if dim2 in lims.keys():
                    ax.set_xlim(lims[dim2])
        return artists

    def init():
        start = 0
        stop = start + results.n_per_sample[start].astype(int_type)
        for i in range(ndims):
            for j in range(ndims):
                axs[i][j].clear()

        artists = []

        artists = _get_artists(artists, start, stop)
        return artists

    def update(start):
        stop = start + results.n_per_sample[start].astype(int_type)
        for i in range(ndims):
            for j in range(ndims):
                axs[i][j].clear()

        artists = []

        artists = _get_artists(artists, start, stop)
        return artists

    ani = FuncAnimation(fig, update, frames=jnp.arange(1, results.num_samples),
                        init_func=init, blit=True)

    ani.save(save_name, fps=results.n_per_sample[0] / 2.)


def add_colorbar_to_axes(ax, cmap, norm=None, vmin=None, vmax=None, label=None):
    """
    Add colorbar to axes easily.

    Args:
        ax: Axes
        cmap: str or cmap
        norm: Normalize or None
        vmin: lower limit of color if norm is None
        vmax: upper limit of color if norm is None
    """
    divider = make_axes_locatable(ax)
    cax = divider.append_axes('right', size='5%', pad=0.05)
    if norm is None:
        norm = plt.Normalize(vmin=vmin, vmax=vmax)
    sm = plt.cm.ScalarMappable(norm, cmap=plt.colormaps.get_cmap(cmap))
    if label is None:
        ax.figure.colorbar(sm, cax=cax, orientation='vertical')
    else:
        ax.figure.colorbar(sm, cax=cax, orientation='vertical', label=label)


def corner_cornerplot(results: NestedSamplerResults):
    try:
        import corner
    except ImportError:
        warnings.warn("You must run `pip install corner`")
        exit(0)
    try:
        import arviz as az
    except ImportError:
        warnings.warn("You must run `pip install arviz`")
        exit(0)
    samples = resample(random.PRNGKey(42), results.samples, results.log_dp_mean, S=int(results.ESS))
    corner.corner(az.from_dict(posterior=jax.tree.map(lambda x: x[None], samples)), )
    plt.show()



from typing import NamedTuple

import jax.numpy as jnp
import numpy as np

from jaxns.internals.namedtuple_utils import serialise_namedtuple, \
    deserialise_namedtuple, issubclass_namedtuple, serialise_ndarray


class MockAge(NamedTuple):
    years: int
    months: int


class MockPerson(NamedTuple):
    name: str
    age: MockAge


def test_isinstance_namedtuple():
    # Example NamedTuple
    data = MockPerson('Alice', MockAge(25, 6))
    assert isinstance(data, MockPerson)

    data = ()
    assert not isinstance(data, MockPerson)


def test_issubclass_namedtuple():
    class TestSubclass(NamedTuple):
        pass

    assert issubclass_namedtuple(TestSubclass)

    class TestSubclass(tuple):
        pass

    assert not issubclass_namedtuple(TestSubclass)

    class C:
        pass

    assert not issubclass_namedtuple(C)


def test_serialise_namedtuple():
    # Example NamedTuple
    data = MockPerson('Alice', MockAge(25, 6))
    # Serialise
    serialized_data = serialise_namedtuple(data)
    print(serialized_data)

    # Deserialize
    restored_data = deserialise_namedtuple(serialized_data)
    print(restored_data)

    assert data == restored_data


def test_serialise_ndarray():
    array = np.array([1, 2, 3])
    serialized_data = serialise_ndarray(array)
    print(serialized_data)
    restored_data = deserialise_namedtuple(serialized_data)
    print(restored_data)

    np.testing.assert_allclose(array, restored_data)


def test_serialise_jax_ndarray():
    array = jnp.array([1, 2, 3])
    serialized_data = serialise_ndarray(array)
    print(serialized_data)
    restored_data = deserialise_namedtuple(serialized_data)
    print(restored_data)

    np.testing.assert_allclose(array, restored_data)



import jax
import numpy as np
from jax import numpy as jnp
from jax._src.scipy.special import logit

from jaxns.internals.constraint_bijections import quick_unit, quick_unit_inverse


def test_quick_unit():
    x = jnp.linspace(-10, 10, 1000000)
    y = quick_unit(x)
    assert np.all(y <= 1)
    assert np.all(y >= 0)
    x_reconstructed = quick_unit_inverse(y)
    np.testing.assert_allclose(x, x_reconstructed, atol=2e-5)

    g = jax.grad(quick_unit)
    assert np.all(np.isfinite(jax.vmap(g)(x)))
    assert np.isfinite(g(0.))

    h = jax.grad(quick_unit_inverse)
    assert np.all(np.isfinite(jax.vmap(h)(y)))
    assert np.isfinite(h(0.5))

    # Test performance against sigmoid and logit
    import time
    for f in [quick_unit, jax.nn.sigmoid]:
        g = jax.jit(f).lower(x).compile()
        t0 = time.time()
        for _ in range(1000):
            g(x).block_until_ready()
        print(f"{f.__name__} {time.time() - t0}s")

    for f in [quick_unit_inverse, logit]:
        g = jax.jit(f).lower(y).compile()
        t0 = time.time()
        for _ in range(1000):
            g(y).block_until_ready()
        print(f"{f.__name__} {time.time() - t0}s")
